//EmysSimulator.cs

/*
    Emys Simulator 
    version 0.8
    
    A simulator for studying the population dynamics of groups of Emys Orbicularis
    Copyright (C) 2009 Aureliano Rama

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA   
*/

/*

	Open a shell with .NET or Mono on the path.
	
	Mono
	mcs EmysSimulator.cs
	mono EmysSimulator.exe 
	
	.NET
	csc EmysSimulator.cs
	EmysSimulator.exe
	
	use "EmysSimulator.exe nnn" to change StartingPopulationCount to nnn 
	without the need to recompile
	

*/

using System;
using System.Collections;

namespace EmysSimulator {

	
	public class Env {
		//Simulation
		public const int RunMaxTime=500;
		public const int SimulationRuns=50;
		public const int PopulationExplosionThreshold = 1000;
		
		//Reproduction
		public const int OvulationProbability=95;
		public const int MeanEggNumber=6;
		public const int NestPredationProbability=90;
		public const int HatchProbability=95;
		public const int ThirdOvulationProbability=10;
		public const int MalesSemiAdultnessAge=6;
		public const int MalesSexualMaturityAge=8;
		public const int FemalesSemiAdultnessAge=10;
		public const int FemalesSexualMaturityAge=12;
		
		//Death
		public const int MeanDeathAge=50;

		//Starting Population
		public const int StartingPopulationCount=50;
		public const int StartingPopulationMinAge = FemalesSexualMaturityAge+1; 
		public const int StartingPopulationMaxAge = MeanDeathAge; 
		
		//Batch Reintroduction
		public const int BatchReintroductionAmount = 5;
		public const int BatchReintroductionDelay = 0;
		public const int BatchReintroductionStep = 5;
		public const int BatchReintroductionTotal = 60;
		public const int BatchReintroductionMinAge = FemalesSexualMaturityAge+1; 
		public const int BatchReintroductionMaxAge = MeanDeathAge; 

		
		//Youth Death Probability
		//Constant population with fluctuations
		public const int NewbornDeathProbability=80;
		public const int Age01DeathProbability=80;
		public const int Age02DeathProbability=80;
		public const int Age03DeathProbability=26;
		public const int Age04DeathProbability=2;
		public const int Age05DeathProbability=0;
		public const int Age06DeathProbability=0;
		public const int Age07DeathProbability=0;
		public const int Age08DeathProbability=0;
		public const int Age09DeathProbability=0;
		public const int Age10DeathProbability=0;
		public const int Age11DeathProbability=0;
		public const int Age12DeathProbability=0;
		
		//Rapid Extinction
		// public const int NewbornDeathProbability=80;
		// public const int Age01DeathProbability=80;
		// public const int Age02DeathProbability=80;
		// public const int Age03DeathProbability=26;
		// public const int Age04DeathProbability=10;
		// public const int Age05DeathProbability=0;
		// public const int Age06DeathProbability=0;
		// public const int Age07DeathProbability=0;
		// public const int Age08DeathProbability=0;
		// public const int Age09DeathProbability=0;
		// public const int Age10DeathProbability=0;
		// public const int Age11DeathProbability=0;
		// public const int Age12DeathProbability=0;

		//Rapid Exponential Explosion
		// public const int NewbornDeathProbability=80;
		// public const int Age01DeathProbability=80;
		// public const int Age02DeathProbability=80;
		// public const int Age03DeathProbability=20;
		// public const int Age04DeathProbability=0;
		// public const int Age05DeathProbability=0;
		// public const int Age06DeathProbability=0;
		// public const int Age07DeathProbability=0;
		// public const int Age08DeathProbability=0;
		// public const int Age09DeathProbability=0;
		// public const int Age10DeathProbability=0;
		// public const int Age11DeathProbability=0;
		// public const int Age12DeathProbability=0;
		
		
		//Even More Rapid Exponential Explosion
		// public const int NewbornDeathProbability=80;
		// public const int Age01DeathProbability=80;
		// public const int Age02DeathProbability=80;
		// public const int Age03DeathProbability=0;
		// public const int Age04DeathProbability=0;
		// public const int Age05DeathProbability=0;
		// public const int Age06DeathProbability=0;
		// public const int Age07DeathProbability=0;
		// public const int Age08DeathProbability=0;
		// public const int Age09DeathProbability=0;
		// public const int Age10DeathProbability=0;
		// public const int Age11DeathProbability=0;
		// public const int Age12DeathProbability=0;

		//Even More Rapid Extinction
		// public const int NewbornDeathProbability=80;
		// public const int Age01DeathProbability=80;
		// public const int Age02DeathProbability=80;
		// public const int Age03DeathProbability=80;
		// public const int Age04DeathProbability=80;
		// public const int Age05DeathProbability=0;
		// public const int Age06DeathProbability=0;
		// public const int Age07DeathProbability=0;
		// public const int Age08DeathProbability=0;
		// public const int Age09DeathProbability=0;
		// public const int Age10DeathProbability=0;
		// public const int Age11DeathProbability=0;
		// public const int Age12DeathProbability=0;
		
		
		
		public static int DeathProbability(int age) {
			int ret;
			switch(age) {
				case 1:
					ret=Age01DeathProbability;
					break;
				case 2:
					ret=Age02DeathProbability;
					break;
				case 3:
					ret=Age03DeathProbability;
					break;
				case 4:
					ret=Age04DeathProbability;
					break;
				case 5:
					ret=Age05DeathProbability;
					break;
				case 6:
					ret=Age06DeathProbability;
					break;
				case 7:
					ret=Age07DeathProbability;
					break;
				case 8:
					ret=Age08DeathProbability;
					break;
				case 9:
					ret=Age09DeathProbability;
					break;
				case 10:
					ret=Age10DeathProbability;
					break;
				case 11:
					ret=Age11DeathProbability;
					break;
				case 12:
					ret=Age12DeathProbability;
					break;
				default:
					ret=0;
					break;
			}
			return ret;
		}
		
	}
	
	public class EmysHashSet {
		Hashtable table;
		
		public EmysHashSet() {
			table = new Hashtable();
		}
		
		public EmysHashSet(int dim) {
			table = new Hashtable(dim);
		}
		
		public void Add(Emys e) {
			//if(!table.Contains(e.Id))
			table.Add(e.Id, e);
		}
		
		public bool Contains(Emys e) {
			return table.Contains(e.Id);
		}
		
		public void Clear() {
			table.Clear();
		}
		
		public void AddAll(EmysHashSet e) {
			IEnumerator en = e.GetEnumerator();
			while(en.MoveNext()) {
				Emys em = (Emys) (en.Current);
				//if(!table.Contains(em.Id))
				table.Add(em.Id, em);
			}
		}
		
		public void RemoveAll(EmysHashSet e) {
			IEnumerator en = e.GetEnumerator();
			while(en.MoveNext()) {
				table.Remove(((Emys) (en.Current)).Id);
			}
		}
		
		public int Count {
			get {
				return table.Count;
			}
		}
		
		public void Remove(Emys e) {
			table.Remove(e.Id);
		}
		
		class EmysEnumerator : IEnumerator {
			IDictionaryEnumerator en;
			
			public EmysEnumerator(EmysHashSet ehs) {
				en = ehs.table.GetEnumerator();
			}
			
			public bool MoveNext() {
				return en.MoveNext();
			}
			
			public void Reset() {
				en.Reset();
			}
			
			public Object Current {
				get {
					return en.Value;
				}
			}
		}

		
		public IEnumerator GetEnumerator() {
			return new EmysEnumerator(this);
		}
	
	}

	public enum Sex {
		MALE,
		FEMALE
	}
	
	public class Tools {
		public static int seed = (int) DateTime.Now.Ticks;
		static Random r = new Random(seed);
		const double sqrt2 = 1.4142135623730950488016887242097;//sqrt(2)

		public static bool GetRandomBool() {
			//return ((r.Next()>>31)==1);
			return ((r.Next() & 128)==0);
		}
		
		public static int GetRandomGauss(int media, int varianza) {
			
			return (int) (((double)media) + ((double)varianza) * GetNormalBMP());
			//return (((r.Next())%(varianza*2)) + media - varianza);
			//return (r.Next())%(media*2);
		}
		
		public static int GetRandomInRange(int low, int max) {
			return ((r.Next())%(max+1-low))+low;			
		}
		
		public static int GetProbability() {
			return GetRandomInRange(0,100);
		}

		public static double GetNormalCLT() {
			//Central Limit Theorem
			double x = 0;
			// the general formula is: (sum(1,n,Xi)-n/2)/sqrt(n/12)
			//so if n==12, this is the result
			// for (int i = 0; i < 12; i++) x += r.NextDouble();
			// return x - 6;
			for (int i = 0; i < 24; i++) x += r.NextDouble();
			return (x - 12)/sqrt2;
		}
		
		static double y1, y2;
		static bool phaseA=true;
		public static double GetNormalBMP() {
			//Box-Muller Polar
			double x1, x2, w;
			
			if(phaseA) {
				phaseA=false;
				do {
					x1 = 2.0 * r.NextDouble() - 1.0;
					x2 = 2.0 * r.NextDouble() - 1.0;
					w = x1 * x1 + x2 * x2;
				} while ( w >= 1.0 );

				w = Math.Sqrt( (-2.0 * Math.Log( w ) ) / w );
				y1 = x1 * w;
				y2 = x2 * w;
				return y1;
			}
			else {
				phaseA=true;
				return y2;
			}
		}
		
	}
	
	public class Emys {
		static int next_id=0;
		int id;
		int age;
		int death;
		Sex sex;
		
		
		public Emys() {
			id=next_id++;
			if(Tools.GetRandomBool())
				sex = Sex.MALE;
			else
				sex = Sex.FEMALE;
			age=0;
			death = Tools.GetRandomGauss(Env.MeanDeathAge-Env.FemalesSexualMaturityAge, 5)+Env.FemalesSexualMaturityAge+1;
		}
		
		public Emys(int age) {
			id=next_id++;
			if(Tools.GetRandomBool())
				sex = Sex.MALE;
			else
				sex = Sex.FEMALE;
			this.age=age;
			death = Tools.GetRandomGauss(Env.MeanDeathAge-Env.FemalesSexualMaturityAge-age, 5)+Env.FemalesSexualMaturityAge+age+1;
		}
		
		public Sex Sex{
			get{
				return sex;
			}
		}
		
		public int Id{
			get{
				return id;
			}
		}
		
		public int Age{
			get{
				return age;
			}
			set{
				age = value; 
			}
    }
		
		public int Death {
			get{
				return death;
			}
		}
		
		public bool isYoungMale() {
			return (sex==Sex.MALE)&&(age<Env.MalesSemiAdultnessAge);
		}
		
		public bool isSemiadultMale() {
			return (sex==Sex.MALE)&&(age>=Env.MalesSemiAdultnessAge)&&(age<Env.MalesSexualMaturityAge);
		}
		
		public bool isAdultMale() {
			return (sex==Sex.MALE)&&(age>=Env.MalesSexualMaturityAge);
		}

		public bool isYoungFemale() {
			return (sex==Sex.FEMALE)&&(age<Env.FemalesSemiAdultnessAge);
		}
		
		public bool isSemiadultFemale() {
			return (sex==Sex.FEMALE)&&(age>=Env.FemalesSemiAdultnessAge)&&(age<Env.FemalesSexualMaturityAge);
		}
		
		public bool isAdultFemale() {
			return (sex==Sex.FEMALE)&&(age>=Env.FemalesSexualMaturityAge);
		}
		
		public bool isYoung() {
			return (isYoungMale() || isYoungFemale());
		}
		
		public bool isSemiadult() {
			return (isSemiadultMale() || isSemiadultFemale());
		}
		
		public bool isAdult() {
			return (isAdultMale() || isAdultFemale());
		}
		
		public EmysHashSet Reproduce() {
			EmysHashSet ret = new EmysHashSet();
			if(isAdultFemale()) {
				int r=Tools.GetProbability();
				if(r<=Env.OvulationProbability) {
					r=Tools.GetProbability();
					if(r<Env.NestPredationProbability) {
						
						int n1 = Tools.GetRandomGauss(Env.MeanEggNumber,2);
						int n2 = Env.HatchProbability;
						if(n2<100)
							n1 = (int) ((((double)n2)/100) * n1);
						
						for(int i=0; i<n1; i++)
							ret.Add(new Emys());
					}
				}
			}
			return ret;
		}
		
		public EmysHashSet Reproduce(int prob) {
			int r=Tools.GetProbability();
			if(r<prob)
				return Reproduce();
			else
				return new EmysHashSet(0);
		}
		
		public bool DieIfYouMust() {
			int prob = Tools.GetProbability();
			if(prob<Env.DeathProbability(age))
				return true;
			else
				return false;
		}
		
	}
	
	
	class Simulator {
		static void Main(String[] args) {
			
			double[][] valori = new double[Env.RunMaxTime][];
			
			for(int i=0; i<Env.RunMaxTime; i++) {
				valori[i] = new double[5]{0,0,0,0,0};
			}
			
			int extinctions = 0;
			int explosions = 0;
			int yearEarliestExplosion = Env.RunMaxTime;
			
			
			int runcount = 0;
			
			int startingPopulation;
			if(args.Length>0)
				startingPopulation = Convert.ToInt32(args[0]);
			else
				startingPopulation = Env.StartingPopulationCount;
			
			if(args.Length>1)
				
			
			while(runcount<Env.SimulationRuns) {
				runcount++;
				int year = 0;
				EmysHashSet population = new EmysHashSet();
				int adults=0;
				int semiadults=0;
				int youngs=0;
				int newborns=0;
				
				const int veryOldAge = 2*Env.MeanDeathAge;
				int[] popcounts = new int[veryOldAge+1];
				for(int i=0; i<veryOldAge+1; i++)
					popcounts[i]=0;

				//Creation of Starting Population
				int eta;
				for(int i=0; i<startingPopulation; i++) {
					eta=Tools.GetRandomInRange(Env.StartingPopulationMinAge,Env.StartingPopulationMaxAge);
					population.Add(new Emys(eta));
					adults++;
					if(eta<veryOldAge)
						popcounts[eta]++;
					else
						popcounts[veryOldAge]++;
				}
				
				//Simulation Cycle
				int prob;
				EmysHashSet nati = new EmysHashSet();
				EmysHashSet morti = new EmysHashSet();
				bool adultMales;
				int reintroduced = 0;
				while(true) {
					year++;
					
					//Batch Reintroduction
					if(reintroduced < Env.BatchReintroductionTotal and year >= Env.BatchReintroductionDelay )
						if(((year-Env.BatchReintroductionDelay) % Env.BatchReintroductionStep) == 0)
							for(int i=0; i<Env.BatchReintroductionAmount; i++) {
								reintroduced++;
								eta=Tools.GetRandomInRange(Env.BatchReintroductionMinAge,Env.BatchReintroductionMaxAge);
								population.Add(new Emys(eta));
								adults++;
								if(eta<veryOldAge)
									popcounts[eta]++;
								else
									popcounts[veryOldAge]++;
							}
							
					//Ovulation and Lay 1;
					adultMales=false; //there MUST be at least one adult male for anything to be born!
					foreach(Emys e in population) {
						nati.AddAll(e.Reproduce());
						adultMales = adultMales || e.isAdultMale();
					}
					if(!adultMales) 
						nati.Clear();
					
					//Predation 1;
					foreach(Emys e in nati) {
						prob = Tools.GetProbability();
						if(prob>Env.NewbornDeathProbability) {
							population.Add(e);
							newborns++;
							popcounts[0]++;
						}
					}
					nati.Clear();
					
					//Ovulation and Lay 2;
					adultMales=false; //there MUST be at least one adult male for anything to be born!
					foreach(Emys e in population) {
						nati.AddAll(e.Reproduce());
						adultMales = adultMales || e.isAdultMale();
					}
					if(!adultMales) 
						nati.Clear();
					
					//Predation 2;
					foreach(Emys e in nati) {
						prob = Tools.GetProbability();
						if(prob>Env.NewbornDeathProbability) {
							population.Add(e);
							newborns++;
							popcounts[0]++;
						}
					}
					nati.Clear();
					
					//Ovulation and Lay 3;
					adultMales=false; //there MUST be at least one adult male for anything to be born!
					foreach(Emys e in population) {
						nati.AddAll(e.Reproduce(Env.ThirdOvulationProbability));
						adultMales = adultMales || e.isAdultMale();
					}
					if(!adultMales)
						nati.Clear();
					
					//Predation 3;
					foreach(Emys e in nati) {
						prob = Tools.GetProbability();
						if(prob>Env.NewbornDeathProbability) {
							population.Add(e);
							newborns++;
							popcounts[0]++;
						}
					}
					nati.Clear();
					
					//Youth Deaths and Death for old age
					foreach(Emys e in population) {
						if(e.DieIfYouMust()){
							morti.Add(e);
							if(e.Age<veryOldAge)
								popcounts[e.Age]--;
							else
								popcounts[veryOldAge]--;
							if(e.Age==0)
								newborns--;
							else if(e.isYoung())
								youngs--;
							else if(e.isAdult())
								adults--;
							else
								semiadults--;
						}
						else if(e.Age>e.Death) {
							morti.Add(e);
							adults--;
							if(e.Age<veryOldAge)
								popcounts[e.Age]--;
							else
								popcounts[veryOldAge]--;
						}
					}
					if(morti.Count>0) {
						population.RemoveAll(morti);
						morti.Clear();
					}
					
					
					valori[year-1][0] += population.Count;
					valori[year-1][1] += newborns;
					valori[year-1][2] += youngs;
					valori[year-1][3] += semiadults;
					valori[year-1][4] += adults;

					//Aging Process
					int oldage, newage;
					foreach(Emys e in population) {
						oldage = (e.Age==0)?0:(e.isAdult()?1:(e.isYoung()?3:2));
						if(e.Age<veryOldAge)
							popcounts[e.Age]--;
						else
							popcounts[veryOldAge]--;

						e.Age = e.Age + 1;

						if(e.Age<veryOldAge)
							popcounts[e.Age]++;
						else
							popcounts[veryOldAge]++;

						newage = e.isAdult()?1:(e.isYoung()?3:2);
						if(newage!=oldage) {
							if(oldage==0) { newborns--; youngs++; }
							else {
								if(oldage==2) semiadults--;
								else youngs--;
								if(newage==1) adults++;
								else semiadults++;
							}
						}
					}
					
					
					//Halting Conditions
					if(population.Count==0) {
						extinctions++;
						break;
					}
					
					if(population.Count>Env.PopulationExplosionThreshold) {
						if(year<yearEarliestExplosion)
							yearEarliestExplosion=year;
						explosions++;
						break;
					}
						
					if(year>=Env.RunMaxTime)
						break;
				}//while(true) //run
			}//while(runcount<Env.SimulationRuns)
			
			Console.WriteLine("Population;Newborns;Youngs;Semi-Adults;Adults");
			for(int i=0; i<Env.RunMaxTime; i++) 
				Console.WriteLine("{0};{1};{2};{3};{4}",valori[i][0]/Env.SimulationRuns,valori[i][1]/Env.SimulationRuns,
				                                        valori[i][2]/Env.SimulationRuns,valori[i][3]/Env.SimulationRuns,
				                                        valori[i][4]/Env.SimulationRuns);
			
			Console.WriteLine("Starting Population Count: {0}", startingPopulation);
			Console.WriteLine("n.run: {2}, n.extinctions: {0}, n.explosions: {1}", extinctions, explosions, Env.SimulationRuns);
			Console.WriteLine("year Earliest Explosion: {0}", yearEarliestExplosion);
			Console.WriteLine("Age03DeathProbability: {0}, Age04DeathProbability: {1}", Env.Age03DeathProbability, Env.Age04DeathProbability);
			Console.WriteLine("Seed: {0}", Tools.seed);
		}//main
	}//class simulator
	
	
}//namespace
